"""This script is used to repair track file names that may be of an incorrect
format. The expected format for track file name is
'TRACK TITLE - ARTIST NAME.mp3'; this format is required in order for this
library to properly identify potential overlap between tracks in a Spotify
playlist(s) and tracks in the beatcloud. This format is also expected for
pre-processing tracks when populating the track title and artist ID3 tags. This
script is intended to be run as a three stage process...

(1) Running this script with just '--fuzz_ratio' will identify "bad" tracks
where the first part of the file name (split on ' - ') does not match
sufficiently with the track's ID3 tag value for track title. You should adjust
the fuzz ratio such that the output is easily readable and you can assert that
the identified tracks are indeed improperly formatted. If there are properly
formatted tracks that are incorrectly identified, you may add them to the
IGNORE_TRACKS set specified at the bottom of this script. If properly formatted
tracks are improperly identified because the ID3 tag for track title is wrong,
then those files should be re-tagged with Picard or MP3Tag or whatever tagging
software you use. They should then be re-uploaded to the beatcloud with
'upload_music' and '--aws_use_date_modified'.

(2) Running this script with '--fuzz_ratio' and '--replace' will first rename
all the "bad" tracks (with the assumption that they are of the format
'ARTIST NAME - TRACK TITLE.mp3') and then remove the old improperly formatted
file from S3 (PLEASE ONLY DO THIS IF YOU ABSOLUTELY KNOW WHAT YOU'RE DOING!).
The script will NOT proceed to upload the new file; since 'aws s3 cp' uploads
one file at a time, I figured it would be better to use the natural parallelism
that 'aws s3 sync' offers and is implicit with using this library's
'upload_music' functionality).

(2.5) After renaming the files in step 2, you can open Rekordbox and click
'File > Display All Missing Files' which will trigger Rekordbox to identify all
the files you renamed. After doing this, click "Collection" and sort by the
left-most column to aggregate all the missing files (those with a '!' symbol in
the left-most column); you can now shift-click all the missing files and create
a new playlist out of them. Name this playlist "Missing Files" or something
unique. Once this playlist is created, click
'File > Export Collection in xml format'.

(3) Running the script with '--xml_path' and '--xml_swap_playlist' will parse
the XML file (the one you created in step 2.5), find the playlist you created
with all the missing files, and fix all the track locations using the same
operation used to rename the files in step 2.

(3.5) Now that you've modified the XML file so the missing files now have the
correct Location field, you can set your import XML to be this one and import
the tracks in this missing files playlist into your collection to relocate them
while retaining the original beatgrid, hot cues, etc.
"""

# pylint: disable=import-error,no-member
from argparse import ArgumentParser
from datetime import datetime
from glob import glob
import json
import logging
import os
from urllib.parse import quote, unquote

from fuzzywuzzy import fuzz
from bs4 import BeautifulSoup
import eyed3

eyed3.log.setLevel("ERROR")


logger = logging.getLogger(__name__)
IGNORE_TRACKS = set(["Scratch Sentence 5.mp3", "Scratch Sentence 1.mp3"])


try:
    import Levenshtein  # pylint: disable=unused-import
except ImportError:
    logger.warning(
        "NOTE: Track similarity can be made faster by running "
        '`pip install "djtools[levenshtein]"`'
    )


def get_bad_tracks(_args):
    """This function globs for mp3 files on '--usb_path' (ignoring those that
    are in the IGNORE_TRACKS set), reads the track title ID3 tag, splits the
    file name on ' - ', and then computes the Levenshtein similarity between
    the two. If the similarity is below '--fuzz_ratio' these files, along with
    the file name prefix and title tag, are printed and the file paths are
    returned as a list of "bad" tracks.

    Args:
        _args (argparser.Namespace): command-line arguments

    Returns:
        list: paths to files whose ID3 title tag has a Levenshtein similarity
              below '--fuzz_ratio' from the first part of the file name
              (split on ' - ')
    """
    usb_path = os.path.join(_args.usb_path, "DJ Music", "**", "*.mp3").replace(
        os.sep, "/"
    )
    files = glob(usb_path, recursive=True)
    _bad_tracks = []
    for _file in files:
        if os.path.basename(_file) in IGNORE_TRACKS:
            continue

        file_title = os.path.basename(_file).split(" - ")[0]
        tag_title = getattr(eyed3.load(_file).tag, "title")
        fuzz_ratio = fuzz.ratio(
            file_title.lower().strip(), tag_title.lower().strip()
        )
        if (
            fuzz_ratio < _args.fuzz_ratio
            and tag_title not in file_title
            and file_title not in tag_title
        ):
            logger.info(
                f"{os.path.basename(_file)}: {file_title} vs. "
                f"{tag_title} = {fuzz_ratio}"
            )
            _bad_tracks.append(_file)

    logger.info(f"{len(_bad_tracks)} bad tracks")

    return _bad_tracks


def replace_tracks(tracks):
    """Renames "bad" tracks by reversing the prefix and suffix part of the file
    name (split on ' - '). The original "bad" files are deleted from S3.

    Args:
        tracks (list): list of "bad" tracks' file paths
    """
    s3_prefix = "s3://dj.beatcloud.com/dj/music/"
    for track in tracks:
        _dir = os.path.dirname(track)
        base_name = os.path.basename(track)
        sub_dir = os.path.basename(_dir)
        name, ext = os.path.splitext(base_name)
        artist, title = name.split(" - ")
        new_base_name = " - ".join([title, artist]) + ext
        new_name = os.path.join(dir, new_base_name).replace(os.sep, "/")
        os.rename(track, new_name)
        dest = os.path.join(s3_prefix, sub_dir, base_name).replace(os.sep, "/")
        rm_cmd = f'aws s3 rm "{dest}"'
        os.system(rm_cmd)


def fix_track_location(xml_path, playlist):
    """This function parses the XML, finds the playlist with missing files,
    finds the tracks in the missing files playlist, and then performs the
    mirror 'artist - title' -> 'title - artist' swap on the 'Location' field
    before writing the modified XML.

    Args:
        xml_path (str): path to XML containing a playlist of missing files
                        (missing because they were renamed)
        playlist (str): name of the playlist containing missing files
    """
    with open(xml_path, "r", encoding="utf-8") as _file:
        soup = BeautifulSoup(_file.read(), "xml")
    lookup = {
        x["TrackID"]: x for x in soup.find_all("TRACK") if x.get("Location")
    }

    try:
        tracks = get_playlist_track_locations(soup, playlist, lookup)
    except LookupError as exc:
        logger.error(exc)
        return

    logger.info(f'{len(tracks)} tracks to repair in playlist "{playlist}"')
    for track in tracks:
        loc = track["Location"]
        dir_name = os.path.dirname(loc)
        base_name = unquote(os.path.basename(loc))
        base_name, ext = os.path.splitext(base_name)
        artist, title = base_name.split(" - ")
        new_base_name = quote(" - ".join([title, artist]) + ext)
        track["Location"] = os.path.join(dir_name, new_base_name).replace(
            os.sep, "/"
        )
        logger.info(
            f"{unquote(loc)} -> "
            + unquote(
                os.path.join(dir_name, new_base_name).replace(os.sep, "/")
            )
        )

    with open(xml_path, mode="wb", encoding=soup.orignal_encoding) as _file:
        _file.write(soup.prettify("utf-8"))


def get_playlist_track_locations(soup, _playlist, lookup):
    """Finds the playlist in the XML and then creates a list of XML tags for
    the missing tracks.

    Args:
        soup (bs4.BeautifulSoup): parsed XML
        _playlist (str): name of playlist in XML which contains missing files
        lookup (dict): map of TrackIDs to XML tags for missing tracks

    Raises:
        LookupError: XML must contain the playlist

    Returns:
        list: XML tags for tracks which appear in the provided playlist
    """
    try:
        playlist = soup.find_all("NODE", {"Name": _playlist})[0]
    except IndexError:
        raise LookupError(f"{_playlist} not found") from LookupError

    return [lookup[x["Key"]] for x in playlist.children if str(x).strip()]


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--usb_path", required=True, type=str, help="path to USB"
    )
    parser.add_argument(
        "--fuzz_ratio",
        type=int,
        default=0,
        help="threshold dissimilarity between ID3 title tag and "
        "supposed title from file name",
    )
    parser.add_argument(
        "--replace",
        action="store_true",
        help="flag to perform file renaming and S3 delete (follow up "
        "with dj_tools upload_music)",
    )
    parser.add_argument(
        "--xml_path", type=str, help='path to "rekordbox.xml" file'
    )
    parser.add_argument(
        "--xml_swap_playlist",
        type=str,
        help="playlist name in which all tracks should have the "
        "title/artist swap performed on their location",
    )
    args = parser.parse_args()

    if args.fuzz_ratio:
        bad_tracks = get_bad_tracks(args)

        if args.replace:
            _file = f"{datetime.now().timestamp()}.json"
            with open(
                os.path.join("bad_tracks", _file).replace(os.sep, "/"),
                "w",
                encoding="utf-8",
            ) as _file:
                json.dump(bad_tracks, _file)
            replace_tracks(bad_tracks)

    if args.xml_swap_playlist:
        fix_track_location(args.xml_path, args.xml_swap_playlist)



"""Testing for the config_formatter module."""

from pydantic import ValidationError

from djtools.configs.config_formatter import BaseConfigFormatter


def test_base_config_formatter_repr():
    """Define a sample subclass of BaseConfigFormatter for testing."""

    class SampleConfig(BaseConfigFormatter):
        """Dummy class."""

        field1: int
        field2: str

    config = SampleConfig(field1=42, field2="test")
    expected_repr = "SampleConfig(\n\tfield1=42\n\tfield2='test'\n)"
    assert repr(config) == expected_repr


def test_base_config_formatter_nested():
    """Define nested configurations."""

    class NestedConfig(BaseConfigFormatter):
        """Dummy class."""

        nested_field: int

    class ParentConfig(BaseConfigFormatter):
        """Dummy class."""

        parent_field: str
        nested: NestedConfig

    nested_config = NestedConfig(nested_field=99)
    parent_config = ParentConfig(parent_field="parent", nested=nested_config)

    expected_repr = (
        "ParentConfig(\n"
        "\tparent_field='parent'\n"
        "\tnested=NestedConfig(\n"
        "\t\tnested_field=99\n"
        "\t)\n"
        ")"
    )
    assert repr(parent_config) == expected_repr


def test_base_config_formatter_invalid():
    """Ensure invalid input raises a ValidationError."""

    class InvalidConfig(BaseConfigFormatter):
        """Dummy class."""

        field: int

    try:
        InvalidConfig(field="not an int")
        assert False, "Expected ValidationError was not raised"
    except ValidationError:
        pass


def test_base_config_formatter_list_of_dicts():
    """Test __repr__ for a field with a list of dictionaries."""

    class ConfigWithListOfDicts(BaseConfigFormatter):
        """Dummy class."""

        items: list[dict]

    config = ConfigWithListOfDicts(
        items=[{"key1": "value1"}, {"key2": "value2"}]
    )
    expected_repr = (
        "ConfigWithListOfDicts(\n"
        "\titems=[\n"
        "\t\t{'key1': 'value1'},\n"
        "\t\t{'key2': 'value2'}\n"
        "\t]\n"
        ")"
    )
    assert repr(config) == expected_repr



"""Testing for the tracks module."""

import pytest

from djtools.collection.base_track import Track


def test_track_raises_type_error():
    """Test Track class."""
    with pytest.raises(
        TypeError,
        match="Can't instantiate abstract class Track",
    ):
        Track()



"""Testing for the collection module."""

import pytest

from djtools.collection.base_collection import Collection


def test_collection_raises_type_error():
    """Test Collection class."""
    with pytest.raises(
        TypeError,
        match="Can't instantiate abstract class Collection",
    ):
        Collection(path="")



"""Testing for the config module."""

from unittest import mock

import pytest
from jinja2 import Template
from pydantic import ValidationError

from djtools.collection.config import (
    CollectionConfig,
    PlaylistConfig,
    PlaylistConfigContent,
)

from ..test_utils import mock_exists, MockOpen


def test_collectionconfig_collection_is_unset_or_missing():
    """Test for the CollectionConfig class."""
    cfg = {"collection_playlists": True, "collection_path": "not/a/real/path"}
    with pytest.raises(
        RuntimeError,
        match="Using the collection package requires the config option "
        "collection_path to be a valid collection path",
    ):
        CollectionConfig(**cfg)


@mock.patch(
    "djtools.collection.config.Path.exists",
    lambda path: mock_exists(
        [
            ("collection_playlists.yaml", False),
        ],
        path,
    ),
)
def test_collectionconfig_no_collection_playlists_config(rekordbox_xml):
    """Test for the CollectionConfig class."""
    cfg = {"collection_playlists": True, "collection_path": rekordbox_xml}
    with pytest.raises(
        RuntimeError,
        match=(
            "collection_playlists.yaml must exist to use the "
            "collection_playlists feature"
        ),
    ):
        CollectionConfig(**cfg)


@mock.patch(
    "djtools.collection.config.Path.exists",
    lambda path: mock_exists(
        [
            ("collection_playlists.yaml", True),
        ],
        path,
    ),
)
@mock.patch(
    "builtins.open",
    MockOpen(
        files=["collection_playlists.yaml"],
        content="tags: invalid-content",
    ).open,
)
def test_collectionconfig_invalid_collection_playlists_config(rekordbox_xml):
    """Test for the CollectionConfig class."""
    cfg = {"collection_playlists": True, "collection_path": rekordbox_xml}
    with pytest.raises(
        RuntimeError,
        match="collection_playlists.yaml must be a valid YAML to use the "
        "collection_playlists feature",
    ):
        CollectionConfig(**cfg)


@mock.patch(
    "djtools.collection.config.Path.exists",
    lambda path: mock_exists(
        [
            ("collection_playlists.yaml", True),
        ],
        path,
    ),
)
def test_collectionconfig_without_template(
    rekordbox_xml, playlist_config_content, playlist_config_obj
):
    """Test for the CollectionConfig class."""
    cfg = {"collection_playlists": True, "collection_path": rekordbox_xml}
    with mock.patch(
        "builtins.open",
        MockOpen(
            files=["collection_playlists.yaml"],
            content=playlist_config_content,
        ).open,
    ):
        config = CollectionConfig(**cfg)
    assert config.playlist_config == playlist_config_obj


@mock.patch(
    "djtools.collection.config.Path.exists",
    lambda path: mock_exists(
        [
            ("collection_playlists.yaml", True),
        ],
        path,
    ),
)
@mock.patch(
    "builtins.open",
    MockOpen(
        files=["collection_playlists.yaml"],
    ).open,
)
@mock.patch("djtools.collection.config.Environment.get_template")
def test_collectionconfig_with_template(mock_get_template, rekordbox_xml):
    """Test for the CollectionConfig class."""
    new_content = ""
    mock_get_template.return_value = Template(new_content)
    cfg = {"collection_playlists": True, "collection_path": rekordbox_xml}
    config = CollectionConfig(**cfg)
    assert config.playlist_config == PlaylistConfig()


@mock.patch("djtools.collection.config.Environment.get_template")
def test_collectionconfig_with_invalid_template(
    mock_get_template, rekordbox_xml
):
    """Test for the CollectionConfig class."""
    mock_template = mock.MagicMock()
    mock_template.render.side_effect = Exception("Render failed")
    mock_get_template.return_value = mock_template

    cfg = {"collection_playlists": True, "collection_path": rekordbox_xml}
    with pytest.raises(RuntimeError):
        CollectionConfig(**cfg)


def test_playlistconfig_example_config(playlist_config):
    """Test for the PlaylistConfig class."""
    PlaylistConfig(**playlist_config)


def test_playlistconfig_minimal_config():
    """Test for the PlaylistConfig class."""
    playlist_config = {"combiner": None, "tags": None}
    PlaylistConfig(**playlist_config)


def test_playlistconfig_is_invalid(playlist_config):
    """Test for the PlaylistConfig class."""
    playlist_config["invalid_key"] = None
    with pytest.raises(ValidationError):
        PlaylistConfig(**playlist_config)


@pytest.mark.parametrize("key", ["tags", "combiner"])
def test_playlistconfigcontent_with_example_config(key, playlist_config):
    """Test for the PlaylistConfigContent class."""
    PlaylistConfigContent(**playlist_config[key])


def test_playlistconfigcontent_with_minimal_config():
    """Test for the PlaylistConfigContent class."""
    content = {"name": "", "playlists": [{"name": "", "playlists": [""]}, ""]}
    PlaylistConfigContent(**content)


def test_playlistconfigcontent_is_invalid():
    """Test for the PlaylistConfigContent class."""
    content = {"invalid": ""}
    with pytest.raises(ValidationError):
        PlaylistConfigContent(**content)



"""Testing for the config module."""

from unittest import mock

import pytest

from djtools.configs.config import BaseConfig, LogLevel
from djtools.spotify.config import SpotifyConfig


@mock.patch(
    "inspect.stack",
    new=mock.Mock(
        return_value=[
            "",
            "",
            ["", "", "", "build_config"],
            ["", "bin/djtools"],
        ],
    ),
)
def test_baseconfig_limits_repr_in_cli_execution():
    """Test for the BaseConfig class."""
    BaseConfig(verbosity=0)


def test_baseconfig_repr_differentiates_baseconfig_and_subclasses():
    """Test for the BaseConfig class."""
    base_config = BaseConfig()
    assert f"log_level={repr(LogLevel.INFO)}" in repr(base_config)
    assert "verbosity=0" in repr(base_config)

    spotify_config = SpotifyConfig()
    assert f"log_level={repr(LogLevel.INFO)}" not in repr(spotify_config)
    assert "verbosity=0" not in repr(spotify_config)


@pytest.mark.parametrize("num_elements", [2, 4, 6])
def test_baseconfig_repr_indents_lists(num_elements):
    """Test for the BaseConfig class."""
    list_elements = [{"name": "thing"}] * num_elements
    spotify_config = SpotifyConfig(spotify_playlist_subreddits=list_elements)
    repr_string = repr(spotify_config)
    assert repr_string.count("\n\t\t") == len(list_elements)



"""Testing for the url_download module."""

# pylint: disable=duplicate-code
from pathlib import Path
from unittest import mock

import pytest

from djtools.utils.url_download import fix_up, url_download


@pytest.mark.parametrize(
    "test_assets",
    [
        (
            "Some Artist - Some Track-1224569284.mp3",
            "Some Track - Some Artist.mp3",
        ),
        (
            "Some Artist-891238 - Some Track-1224569284.wav",
            "Some Track - Some Artist-891238.wav",
        ),
        (
            "Some Artist - Some Track.mp3",
            "Some Track - Some Artist.mp3",
        ),
    ],
)
def test_fix_up(test_assets):
    """Test for the fix_up function."""
    test_file, expected_clean_file = map(Path, test_assets)
    clean_file = fix_up(test_file)
    assert clean_file == expected_clean_file


def test_url_download(tmpdir, config):
    """Test for the url_download function."""
    tmpdir = Path(tmpdir)
    config.utils.url_download = (
        "https://soundcloud.com/aweeeezy_music/sets/test-download"
    )
    config.utils.audio_destination = tmpdir / "new_dir"
    assert not config.utils.audio_destination.exists()

    def dummy_func():
        with open(
            config.utils.audio_destination / "file.mp3",
            mode="w",
            encoding="utf-8",
        ) as _file:
            _file.write("")

    with mock.patch(
        "youtube_dl.YoutubeDL",
    ) as mock_ytdl:
        context = mock_ytdl.return_value.__enter__.return_value
        context.download.side_effect = lambda *args, **kwargs: dummy_func()
        url_download(config)

    assert config.utils.audio_destination.exists()
    assert len(list(config.utils.audio_destination.iterdir())) == 1



"""Testing for the helpers module."""

import logging
from datetime import datetime
from pathlib import Path
from typing import Optional
from unittest import mock

import pytest
from pydub import AudioSegment, generators

from djtools.utils.config import TrimInitialSilenceMode
from djtools.utils.helpers import (
    compute_distance,
    find_matches,
    get_beatcloud_tracks,
    get_local_tracks,
    get_playlist_tracks,
    get_spotify_tracks,
    initialize_logger,
    make_path,
    process_parallel,
    reverse_title_and_artist,
    trim_initial_silence,
)


@pytest.mark.parametrize("track_a", ["some track", "another track"])
@pytest.mark.parametrize("track_b", ["some track", "another track"])
def test_compute_distance(track_a, track_b):
    """Test for the compute_distance function."""
    ret = compute_distance(
        "playlist",
        track_a,
        track_b,
        100,
    )
    if track_a == track_b:
        assert ret
        assert ret[0] == "playlist"
        assert ret[1] == track_a
        assert ret[2] == track_b
        assert ret[3] == 100
    else:
        assert not ret


def test_find_matches(config):
    """Test for the find_matches function."""
    config.utils.check_tracks_fuzz_ratio = 100
    expected_matches = [
        "track 1 - someone unique",
        "track 2 - seen them before",
    ]
    matches = find_matches(
        compare_tracks={
            "playlist_a": expected_matches,
            "playlist_b": [
                "track 3 - special person",
                "track 4 - seen them before",
            ],
        },
        beatcloud_tracks=[
            "track 5 - who's that?",
        ]
        + expected_matches,
        config=config,
    )
    assert all(
        match[-1] == config.utils.check_tracks_fuzz_ratio for match in matches
    )
    assert len(matches) == 2
    assert {x[1] for x in matches} == set(expected_matches)


@pytest.mark.parametrize(
    "proc_dump",
    [
        [],
        [
            "aweeeezy/Bass/2022-12-21/track - artist.mp3",
            "aweeeezy/Techno/2022-12-21/track - artist.mp3",
        ],
    ],
)
@mock.patch("djtools.utils.helpers.check_output")
def test_get_beatcloud_tracks(mock_os_popen, proc_dump):
    """Test for the get_beatcloud_tracks function."""
    bucket_url = "s3://some-bucket.com"
    proc_dump = list(map(Path, proc_dump))
    mock_os_popen.return_value = b"\n".join(
        map(lambda x: x.as_posix().encode(), proc_dump)
    )
    tracks = get_beatcloud_tracks(bucket_url)
    mock_os_popen.assert_called_once()
    assert len(tracks) == len(proc_dump)
    for track, line in zip(tracks, proc_dump):
        assert track == line


def test_get_local_tracks_dir_does_not_exist(config, caplog):
    """Test for the get_local_tracks function."""
    caplog.set_level("INFO")
    local_dir = Path("not-a-real-dir")
    config.utils.local_dirs = [local_dir]
    local_dir_tracks = get_local_tracks(config)
    assert caplog.records[0].message == (
        f"{local_dir.as_posix()} does not exist; will not be able to check "
        "its contents against the beatcloud"
    )
    assert not local_dir_tracks
    assert caplog.records[1].message == "Got 0 files under local directories"


def test_get_local_tracks_dir_does_exist(tmpdir, config, caplog):
    """Test for the get_local_tracks function."""
    caplog.set_level("INFO")
    check_dirs = []
    tmpdir = Path(tmpdir)
    for _dir in ["dir1", "dir2"]:
        path = tmpdir / _dir
        path.mkdir(parents=True, exist_ok=True)
        check_dirs.append(path)
    config.utils.local_dirs = check_dirs
    beatcloud_tracks = ["test_file1.mp3", "test_file2.mp3"]
    for index, track in enumerate(beatcloud_tracks):
        with open(
            check_dirs[index % len(check_dirs)] / f"{track}",
            mode="w",
            encoding="utf-8",
        ) as _file:
            _file.write("")
    local_dir_tracks = get_local_tracks(config)
    assert set(local_dir_tracks).union(set(check_dirs)) == set(check_dirs)
    assert (
        caplog.records[0].message
        == f"Got {len(check_dirs)} files under local directories"
    )


def test_get_playlist_tracks():
    """Test for the get_playlist_tracks function."""
    mock_client = mock.MagicMock()
    mock_client.playlist.return_value = {
        "tracks": {
            "items": [
                {
                    "track": {
                        "name": "track title",
                        "artists": [{"name": "artist name"}],
                    },
                },
                {
                    "track": {
                        "name": "another track title",
                        "artists": [
                            {"name": "another artist name"},
                            {"name": "a second artist name"},
                        ],
                    },
                },
            ],
            "next": True,
        },
    }
    mock_client.next.return_value = {
        "items": [
            {
                "track": {
                    "name": "last track title",
                    "artists": [{"name": "final artist name"}],
                },
            },
        ],
        "next": False,
    }

    tracks = get_playlist_tracks(mock_client, "some ID")

    expected = list(mock_client.playlist.return_value["tracks"]["items"])
    expected.extend(list(mock_client.next.return_value["items"]))
    assert tracks == expected


def test_get_playlist_tracks_handles_exception():
    """Test for the get_playlist_tracks function."""
    mock_client = mock.MagicMock()
    mock_client.playlist.side_effect = Exception("API Error")
    test_playlist_id = "some ID"

    with pytest.raises(
        RuntimeError,
        match=f"Failed to get playlist with ID {test_playlist_id}",
    ):
        get_playlist_tracks(mock_client, test_playlist_id)


@mock.patch("djtools.utils.helpers.get_spotify_client", new=mock.Mock())
@mock.patch(
    "djtools.utils.helpers.get_playlist_ids",
    new=mock.Mock(
        return_value={"r/techno | Top weekly Posts": "5gex4eBgWH9nieoVuV8hDC"}
    ),
)
def test_get_spotify_tracks_no_matching_playlist_id(config, caplog):
    """Test for the get_spotify_tracks function."""
    caplog.set_level("ERROR")
    playlist = "not-a-real-playlist"
    get_spotify_tracks(config, [playlist])
    assert (
        caplog.records[0].message
        == f"{playlist} not in spotify_playlists.yaml"
    )


@pytest.mark.parametrize("verbosity", [0, 1])
@mock.patch("djtools.utils.helpers.get_spotify_client", new=mock.Mock())
@mock.patch(
    "djtools.utils.helpers.get_playlist_ids",
    new=mock.Mock(
        return_value={"r/techno | Top weekly Posts": "5gex4eBgWH9nieoVuV8hDC"}
    ),
)
@mock.patch(
    "djtools.utils.helpers.get_playlist_tracks",
    return_value={"some track - some artist"},
)
def test_get_spotify_tracks(
    mock_get_playlist_tracks, config, verbosity, caplog
):
    """Test for the get_spotify_tracks function."""
    caplog.set_level("INFO")
    config.verbosity = verbosity
    tracks = get_spotify_tracks(config, ["r/techno | Top weekly Posts"])
    assert isinstance(tracks, dict)
    assert caplog.records[0].message == (
        'Got 1 track from Spotify playlist "r/techno | Top weekly Posts"'
    )
    if verbosity:
        assert caplog.records[1].message == "\tsome track - some artist"
    mock_get_playlist_tracks.assert_called_once()


def test_initialize_logger():
    """Test for the intitialize_logger function."""
    today = f'{datetime.now().strftime("%Y-%m-%d")}.log'
    logger, log_file = initialize_logger()
    assert isinstance(logger, logging.Logger)
    assert log_file.name == today


@pytest.mark.parametrize(
    "kwargs, expected_str_kwarg, expected_path_kwarg",
    [
        ({"str_kwarg": "string kwarg", "path_kwarg": "path kwarg"}, str, Path),
        ({}, type(None), type(None)),
    ],
)
def test_make_path_decorator(kwargs, expected_str_kwarg, expected_path_kwarg):
    """Test for the make_path decorator function."""

    @make_path
    def foo(  # pylint: disable=disallowed-name
        str_arg: str,
        path_arg: Path,
        str_kwarg: Optional[str] = None,
        path_kwarg: Optional[Path] = None,
    ):
        assert isinstance(str_arg, str)
        assert isinstance(path_arg, Path)
        assert isinstance(str_kwarg, expected_str_kwarg)
        assert isinstance(path_kwarg, expected_path_kwarg)

    foo("a string arg", "a string arg path", **kwargs)


@pytest.mark.parametrize(
    "arg, kwarg, expected",
    [
        (1, "", 'Error creating Path in function "foo" from positional'),
        ("", 1, 'Error creating Path in function "foo" from keyword'),
    ],
)
def test_make_path_decorator_raises_error(arg, kwarg, expected):
    """Test for the make_path decorator function."""

    @make_path
    def foo(
        path_arg: Path, path_kwarg: Path
    ):  # pylint: disable=disallowed-name
        assert isinstance(path_arg, Path)
        assert isinstance(path_kwarg, Path)

    with pytest.raises(
        RuntimeError,
        match=expected,
    ):
        foo(arg, path_kwarg=kwarg)


@mock.patch("djtools.utils.helpers.AudioSegment.export", new=mock.Mock())
@mock.patch("djtools.utils.helpers.effects.normalize")
def test_process_parallel(mock_normalize, config, audio_file, tmpdir):
    """Test for the process_parallel function."""
    audio, _ = audio_file
    track = {"artist": "Artist", "title": "Title"}
    process_parallel(config, audio, track, tmpdir)
    assert mock_normalize.call_count == 1


def test_reverse_title_and_artist():
    """Test for the reverse_title_and_artist function."""
    path_lookup = {
        "title - artist": "path/to/title - artist.mp3",
        "multiple - hyphens - artist": "path/to/multiple - hyphens - artist.mp3",
    }
    expected = {
        "artist - title": "path/to/title - artist.mp3",
        "artist - multiple - hyphens": "path/to/multiple - hyphens - artist.mp3",
    }
    new_path_lookup = reverse_title_and_artist(path_lookup)
    assert new_path_lookup == expected


@pytest.mark.parametrize(
    "trim_amount",
    [
        -1000,
        0,
        1000,
        TrimInitialSilenceMode.AUTO,
        TrimInitialSilenceMode.SMART,
    ],
)
def test_trim_initial_silence(trim_amount):
    """Test for the trim_initial_silence function."""
    leading_silence = 4567
    step_size = 100
    silence_len = 500
    track_durations = [1234, 2345, 3456]

    # Build up a mock recording of multiple tracks.
    audio = AudioSegment.silent(duration=leading_silence)
    for dur in track_durations:
        audio += generators.WhiteNoise().to_audio_segment(duration=dur)
        audio += AudioSegment.silent(duration=silence_len)
    init_len = len(audio)

    # Track durations have to account for any silence inserted during playback.
    track_durations = [dur + silence_len for dur in track_durations]

    # Sometimes pydub has an off-by-one difference in the length of audio
    # constructed in this way.
    assert abs(init_len - (leading_silence + sum(track_durations))) <= 1

    # Trim leading silence.
    audio = trim_initial_silence(
        audio, track_durations, trim_amount, step_size=step_size
    )

    # The difference between the audio length, before and after, must be
    # approximately the leading silence amount minus the tail silence.
    diff = init_len - len(audio)
    if isinstance(trim_amount, int):
        assert abs(diff - trim_amount) <= 1
    elif trim_amount == TrimInitialSilenceMode.AUTO:
        assert abs(diff - leading_silence) <= 1
    else:
        assert abs(leading_silence - silence_len - diff) <= step_size * 2



"""Testing for the cli_args module."""

from pathlib import Path
from typing import List

import pytest
from pydantic import BaseModel

from djtools.configs.cli_args import (
    _convert_to_paths,
    _parse_json,
    _parse_trim_initial_silence,
    get_arg_parser,
    NonEmptyListElementAction,
)
from djtools.utils.config import TrimInitialSilenceMode


@pytest.mark.parametrize("paths", ["path", ["path1", "path2"]])
def test_convert_to_paths(paths):
    """Test for the convert_to_paths function."""
    paths = _convert_to_paths(paths)
    if isinstance(paths, List):
        for path in paths:
            assert isinstance(path, Path)
    else:
        assert isinstance(paths, Path)


def test_get_arg_parser_arg_for_every_field(config):
    """Test the get_arg_parser function."""
    # Get CLI args.
    parser = get_arg_parser()
    args_set = {
        arg
        for args in [
            ["collection", "--collection-playlists"],
            ["spotify", "--spotify-playlists"],
            ["sync", "--upload-collection"],
            ["utils", "--normalize-audio"],
        ]
        for arg in parser.parse_args(args).__dict__.keys()
    }
    config_set = set()
    for field_name, field_info in config.model_fields.items():
        if isinstance(field_info.annotation, type) and issubclass(
            field_info.annotation, BaseModel
        ):
            sub_model = getattr(config, field_name)
            config_set.update(list(sub_model.model_fields))
            continue

        config_set.add(field_name)

    # Test that the only CLI exclusive args are linking configs and displaying
    # the version.
    cli_only = args_set.difference(config_set)
    expected_cli_only = {"link_configs", "version"}
    assert (
        cli_only == expected_cli_only
    ), f"Expected CLI args to be {expected_cli_only} but got {cli_only}"

    # Test that every config option has a corresponding CLI arg.
    config_only = config_set.difference(args_set)

    # NOTE: playlist_config is the only BaseConfig attribute that doesn't have
    # a corresponding CLI arg. This is because the CollectionConfig preloads
    # the user's collection playlists.
    assert config_only == {"playlist_config"}, (
        "Expected a CLI arg for every config option but these were missing:"
        f"\n{config_only}"
    )


def test_non_empty_list_element_action(namespace):
    """Test for the NonEmptyListElementAction class."""
    namespace.local_dirs = ["first"]
    parser = get_arg_parser()
    action = NonEmptyListElementAction(option_strings=None, dest="local_dirs")
    action(parser, namespace, values=[None, "second"])
    assert namespace.local_dirs == ["first", "second"]


def test_parse_json():
    """Test for the parse_json function."""
    json_string = '{"name": ["stuff"]}'
    json_obj = _parse_json(json_string)
    assert isinstance(json_obj, dict)
    assert isinstance(json_obj["name"], list)


def test_parse_json_invalid():
    """Test for the parse_json function."""
    json_string = '{"name": {"stuff"}}'
    with pytest.raises(ValueError):
        _parse_json(json_string)


def test_parse_trim_initial_silence_int():
    """Test for the _parse_trim_initial_silence function."""
    arg = "1"
    assert isinstance(arg, str)
    result = _parse_trim_initial_silence(arg)
    assert isinstance(result, int)
    assert result == 1


def test_parse_trim_initial_silence_str():
    """Test for the _parse_trim_initial_silence function."""
    arg = TrimInitialSilenceMode.AUTO
    assert isinstance(arg, TrimInitialSilenceMode)
    result = _parse_trim_initial_silence(arg)
    assert isinstance(result, TrimInitialSilenceMode)
    assert result == TrimInitialSilenceMode.AUTO


def test_parse_trim_initial_silence_invalid_str():
    """Test for the _parse_trim_initial_silence function."""
    arg = "stuff"
    with pytest.raises(
        ValueError,
        match=(
            '--trim-initial-silence must be either "auto", "smart", or an '
            "integer."
        ),
    ):
        _parse_trim_initial_silence(arg)



"""Script used to relocate all music files from the old genre-top-level
structure to the new username-top-level structure.
"""

# pylint: disable=too-many-arguments,duplicate-code,no-member
from argparse import ArgumentParser
from concurrent.futures import ThreadPoolExecutor
from itertools import groupby
import json
import logging
import logging.config
from operator import itemgetter
import os
from urllib.parse import quote, unquote

from bs4 import BeautifulSoup
from fuzzywuzzy import fuzz
import spotipy
from spotipy.oauth2 import SpotifyOAuth

for logger in ["spotipy", "urllib3"]:
    logger = logging.getLogger(logger)
    logger.setLevel(logging.CRITICAL)

parent = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
log_conf = os.path.join(
    parent, "src", "djtools", "configs", "logging.conf"
).replace(os.sep, "/")
logging.config.fileConfig(
    fname=log_conf,
    disable_existing_loggers=False,
    defaults={"logfilename": "move_music_new_structure.log"},
)
logger = logging.getLogger(__name__)


def get_spotify_tracks(_config, spotify_playlists):
    """Aggregates the tracks from one or more Spotify playlists into a
    dictionary mapped with track title and artist names.

    Args:
        _config (dict): configuration object
        spotify_playlists (dict): playlist names and IDs

    Returns:
        (dict): Spotify tracks keyed by titles and artist names
    """

    spotify = spotipy.Spotify(
        auth_manager=SpotifyOAuth(
            client_id=_config["spotify_client_id"],
            client_secret=_config["spotify_client_secret"],
            redirect_uri=_config["spotify_redirect_uri"],
            scope="playlist-modify-public",
        )
    )
    _tracks = {}
    for playlist, playlist_id in spotify_playlists.items():
        logger.info(f'Getting tracks from Spotify playlist "{playlist}"...')
        __tracks = get_playlist_tracks(spotify, playlist_id, playlist)
        logger.info(f"Got {len(__tracks)} tracks from {playlist}")
        _tracks.update(__tracks)

    logger.info(f"Got {len(_tracks)} tracks in total")

    return _tracks


def get_playlist_tracks(spotify, playlist_id, _playlist):
    """Queries Spotify API for a playlist and pulls tracks from it.

    Args:
        spotify (spotipy.Spotify): Spotify client
        playlist_id (str): playlist ID of Spotify playlist to pull tracks from

    Raises:
        RuntimeError: playlist_id must correspond with a valid Spotify playlist

    Returns:
        set: Spotify track titles and artist names from a given playlist
    """
    try:
        playlist = spotify.playlist(playlist_id)
    except Exception:
        raise RuntimeError(
            f"Failed to get playlist with ID {playlist_id}"
        ) from Exception

    result = playlist["tracks"]
    _tracks = add_tracks(result, _playlist)

    while result["next"]:
        result = spotify.next(result)
        _tracks.update(add_tracks(result, _playlist))

    return _tracks


def add_tracks(result, playlist):
    """Parses a page of Spotify API result tracks and returns a list of the
    track titles and artist names.

    Args:
        result (spotipy.Tracks): paged result of Spotify tracks

    Returns:
        (list): Spotify track titles and artist names
    """
    _tracks = {}
    for track in result["items"]:
        title = track["track"]["name"]
        artists = ", ".join([y["name"] for y in track["track"]["artists"]])
        _tracks[f"{title} - {artists}"] = {
            "added_at": track["added_at"],
            "added_by": track["added_by"]["id"],
            "playlist": playlist,
        }

    return _tracks


def get_beatcloud_tracks():
    """Lists all the music files in S3 and parses out the track titles and
    artist names.

    Returns:
        list: beatcloud track titles and artist names
    """
    logger.info("Getting tracks from the beatcloud...")
    cmd = "aws s3 ls --recursive s3://dj.beatcloud.com/dj/music/"
    with os.popen(cmd) as proc:
        output = proc.read().split("\n")
    _tracks = [track.split("dj/music/")[-1] for track in output if track]
    logger.info(f"Got {len(_tracks)} tracks")

    return _tracks


def analyze_tracks(_tracks, users):
    """Display Spotify tracks grouped by user and then playlists.

    Args:
        _tracks (dict): map Spotify track - artist -> {added_by, playlist}
        users (dict): map Spotify username -> beatcloud username
    """
    logger.info("Analyzing user contributions to spotify playlists...")
    for user_group_id, user_group in groupby(
        sorted(_tracks.values(), key=itemgetter("added_by")),
        key=itemgetter("added_by"),
    ):
        logger.info(f"{users[user_group_id]} ({user_group_id}):")
        for playlist_group_id, playlist_group in groupby(
            sorted(user_group, key=itemgetter("playlist")),
            key=itemgetter("playlist"),
        ):
            playlist_group = list(playlist_group)
            logger.info(f"\t{playlist_group_id}: {len(playlist_group)}")


def find_local_files(usb_path, remote_files):
    """Validates that each remote file exists locally. Returns the full paths
    of just the validated ones.

    Args:
        usb_path (str): full path to root of USB drive with "DJ Music"
        remote_files (list): list of `beatcloud` files relative to "/dj/music/"

    Returns:
        list: paths to local files that are also in the `beatcloud`
    """
    payload = [[usb_path] * len(remote_files), remote_files]
    with ThreadPoolExecutor(max_workers=os.cpu_count() * 4) as executor:
        _files = list(
            filter(None, list(executor.map(exists_process, *payload)))
        )
    logger.info(f"Found {len(_files)} local files")

    return _files


def exists_process(path, _file):
    """Threaded process that returns the local path, if it exists, to a track
    that exists in the `beatcloud`.

    Args:
        path (str): usb_path
        _file (str): sub-path to beatcloud track (relative to "/dj/music/")

    Returns:
        str: local path to track that exists in the `beatcloud`
    """
    _path = os.path.join(path, "DJ Music", _file).replace(os.sep, "/")

    return _file if os.path.exists(_path) else None


def fix_files(bad_files, _local_files, move_local_files, usb_path, verbosity):
    """Renames local files to fix typos and improve fuzz match results so all
    correct matches happen above the final fuzz_ratio.

    Args:
        bad_files (dict): map bad file names to their corrections
        _local_files (list): list of local file paths
        move_local_files (bool): flag to trigger moving local files
        usb_path (str): path to root of USB containing "DJ Music"
        verbosity (int): verbosity level

    Returns:
        list: _local_files after swapping bad files for their corrections
    """
    logger.info(f'Renaming {len(bad_files)} "bad files"...')
    for bad, good in bad_files.items():
        _bad = os.path.join(usb_path, "DJ Music", bad).replace(os.sep, "/")
        _good = os.path.join(usb_path, "DJ Music", good).replace(os.sep, "/")
        if verbosity > 0:
            logger.info(f"\t{_bad} -> {_good}")
        if move_local_files:
            try:
                os.rename(_bad, _good)
            except Exception as exc:
                logger.error(f'failed to rename "{_bad}" to "{_good}": {exc}')
                continue
        try:
            _index = _local_files.index(bad)
            _local_files[_index] = good
        except Exception:
            continue

    return _local_files


def match_local_files(
    _local_files,
    _tracks,
    fuzz_ratio,
    verbosity,
    ignore,
    _matches,
    cache_fuzz_results,
):
    """Matches local files to Spotify tracks:
        (1) first pass does direct string equality checks between file names
            and `title - artist` strings of Spotify tracks
        (2) second pass fuzzy matches files against Spotify tracks...

    If run with `--cache_fuzz_results`, previously confirmed matches are
    eliminated from consideration during fuzzy matching; this is so it can be
    easier to dial in the proper `fuzz_ratio` while ensuring every matched
    track is correct (and also identify what `bad_files` can be corrected to
    improve matching). The final run should NOT use cached fuzz results.

    Args:
        _local_files (list): list of local file paths
        _tracks (dict): map Spotify track - artist -> {added_by, playlist}
        fuzz_ratio (int): min fuzz ratio to accept file -> Spotify track match
        verbosity (int): verbosity level
        ignore (list): true non-matches with fuzz ratio above "fuzz_ratio"
        _matches (dict): temporary lookup that functions like "ignore"
        cache_fuzz_results (bool): whether or not to populate the fuzz results
                                   cache and ignore results already present in
                                   said cache

    Returns:
        tuple: (map: file path -> Spotify track, list: files not mapped)
    """
    # filter local files by those that can directly be mapped to one of the
    # spotify tracks; also remove the corresponding Spotify tracks from future
    # consideration
    found_tracks = {}
    __local_files = []
    for _file in _local_files:
        name = os.path.splitext(os.path.basename(_file))[0]
        track = _tracks.get(name)
        if not track:
            __local_files.append(_file)
            continue
        del _tracks[name]
        found_tracks[_file] = track
    _local_files = __local_files
    directly_found = len(found_tracks)
    logger.info(
        f"Found {directly_found} tracks...fuzzy searching for "
        f"the remaining {len(_local_files)}"
    )

    # Load cached fuzzy search results...
    # This tool is intended to be used in iterations with progressively lower
    # minimum 'fuzz_ratio' so as to:
    #    - first fix local files with typos or misformattings (correction
    #      mapping in data['bad_files'])
    #    - then verify that every local track pairs properly with the most
    #      similar Spotify track
    #    - once invalid matches start showing up in the results, make sure each
    #      one is definitely not traceable to a Spotify playlist and then add
    #      to data['ignore']
    #
    # Each iteration should actually be a two-stage process:
    #    (1) see results at newly lowered 'fuzz_ratio', using
    #        `.fuzz_cache.json` to filter previous results, and make any
    #        necessary additions to data['bad_files'] and / or data['ignore']
    #    (2) restore `.fuzz_cache.json` to the backup you made before (1) and
    #        confirm the output at the same 'fuzz_ratio' is as expected...
    #
    # Once all results are definitely not a match at 'fuzz_ratio=0', then
    # you've confirmed every local file that can be attributed to a recognized
    # user of data['users']; what remains outside of the username-based folders
    # came from outside of data['spotify_playlists'] or else is an artifact
    # from a time when the beatcloud contained incorrectly named files that had
    # since been corrected.
    if cache_fuzz_results and os.path.exists(".fuzz_cache.json"):
        with open(".fuzz_cache.json", encoding="utf-8") as _file:
            cached_fuzz_results = json.load(_file)
        logger.info(
            f"Ignoring {len(cached_fuzz_results)} fuzz results "
            "previously matched"
        )
        prev_track_count = len(_tracks)
        prev_matches = set(cached_fuzz_results.values())
        _tracks = {k: v for k, v in _tracks.items() if k not in prev_matches}
        logger.info(
            f"Reduced number of Spotify tracks being considered "
            f"from {prev_track_count} to {len(_tracks)}"
        )
    else:
        cached_fuzz_results = {}

    # distribute fuzzy search of a file against every Spotify track...keep the
    # most similar result above 'fuzz_ratio' and add it to the collection of
    # directly matched files -> tracks
    not_matched = []
    ignore = {os.path.splitext(os.path.basename(x))[0] for x in ignore}
    for _file in _local_files:
        name = os.path.splitext(os.path.basename(_file))[0]
        if any((name in x for x in [cached_fuzz_results, ignore, _matches])):
            continue
        payload = [
            [name] * len(_tracks),
            list(_tracks.keys()),
            [fuzz_ratio] * len(_tracks),
        ]
        with ThreadPoolExecutor(max_workers=os.cpu_count() * 4) as executor:
            __matches = sorted(
                filter(None, executor.map(fuzz_process, *payload)),
                reverse=True,
                key=itemgetter(1),
            )
        if not __matches:
            not_matched.append(_file)
            continue
        match, ratio = __matches[0]
        found_tracks[_file] = _tracks[match]
        found_tracks[_file]["fuzz_ratio"] = ratio
        found_tracks[_file]["match"] = match
        # del _tracks[match]
        if cache_fuzz_results:
            cached_fuzz_results[name] = match

    if cache_fuzz_results:
        with open(".fuzz_cache.json", "w", encoding="utf-8") as _file:
            json.dump(cached_fuzz_results, _file)

    # optionally display all the results of the fuzzy search
    if verbosity > 0:
        logger.info("Fuzzy matched files:")
        for name, track in found_tracks.items():
            if not track.get("fuzz_ratio"):
                continue
            logger.info(
                f"\t{track['fuzz_ratio']}: "
                + os.path.splitext(os.path.basename(name))[0]
            )
            logger.info(f"\t    {track['match']}")
    logger.info(f"Fuzzy matched {len(found_tracks) - directly_found} files")
    logger.info(
        f"Unable to find {len(not_matched)} tracks (plus the "
        f"{len(ignore)} in data['ignore'])"
    )

    return found_tracks, not_matched


def move_files(
    found_tracks,
    not_matched,
    users,
    playlist_genres,
    usb_path,  # pylint: disable=too-many-arguments,too-many-locals
    move_local_files,
    not_matched_lookup,
    move_remote_files,
    verbosity,
    ignore,
    bad_files_inverse_lookup,
    xml_path,
    _user_names,
    write_xml,
):
    """Moves files to their new location...
        (1) direct matches and fuzzy match results are moved locally
        (2) not matched local files are moved based on their previous directory
        (3) S3 files are moved from former location to new location

    Args:
        found_tracks (dict): map file path -> Spotify track
        not_matched (list): files not matched to Spotify track
        users (dict): map Spotify username -> beatcloud username
        playlist_genres (dict): map playlist name to genre folder
        usb_path (str): full path to root of USB drive with "DJ Music"
        move_local_files (bool): flag to trigger moving local files
        not_matched_lookup (dict): maps local directories to genre folder
        move_remote_files (bool): flag to trigger moving s3 files
        verbosity (int): verbosity level
        ignore (list): true non-matches with fuzz ratio above "fuzz_ratio"
        bad_files_inverse_lookup (dict): reverse lookup of corrected
                                         "bad_files"...used to move incorrect
                                         version of file in S3 to new corrected
                                         location
        xml_path (str): path to Rekordbox XML; will be rewritten so tracks
                        point to their new locations
    """
    # move every local file matched with a playlist and user to that user's
    # corresponding data['playlist_genres'] folder
    s3_prefix = "s3://dj.beatcloud.com/dj/music/"
    s3_moves = []
    logger.info("Moving matched local files to their new locations...")
    for name, track in found_tracks.items():
        user = users[track["added_by"]]
        genre = playlist_genres[track["playlist"]]
        base_path = os.path.join(user, genre, "old", os.path.basename(name))
        dest_path = os.path.join(usb_path, "DJ Music", base_path).replace(
            os.sep, "/"
        )
        dest_dir = os.path.dirname(dest_path)
        if not os.path.exists(dest_dir):
            if move_local_files:
                make_dirs(dest_dir)
        src_path = os.path.join(usb_path, "DJ Music", name).replace(
            os.sep, "/"
        )
        msg = f"\t{src_path} -> {dest_path}"
        if verbosity > 0:
            logger.info(msg)
        if move_local_files:
            try:
                os.rename(src_path, dest_path)
            except Exception as exc:
                logger.error(
                    f'failed to move matched file "{src_path}" to '
                    f'"{dest_path}": {exc}'
                )
                continue

        # reformat paths for S3 file relocations
        name = bad_files_inverse_lookup.get(name, name)
        src_path = os.path.join(s3_prefix, name).replace(os.sep, "/")
        dest_path = os.path.join(s3_prefix, base_path).replace(os.sep, "/")
        s3_moves.append((src_path, dest_path))

    # move remaining tracks using the original file location's parent directory
    # to determine the proper new location
    logger.info(
        "Moving unmatched local files to the proper location based "
        "on the directory of their previous location..."
    )
    for name in not_matched + ignore:
        parent_dir = os.path.basename(os.path.dirname(name))
        base_path = os.path.join(
            not_matched_lookup[parent_dir], "old", os.path.basename(name)
        )
        dest_path = os.path.join(usb_path, "DJ Music", base_path).replace(
            os.sep, "/"
        )
        dest_dir = os.path.dirname(dest_path)
        if not os.path.exists(dest_dir):
            if move_local_files:
                make_dirs(dest_dir)
        src_path = os.path.join(usb_path, "DJ Music", name).replace(
            os.sep, "/"
        )
        msg = f"\t{src_path} -> {dest_path}"
        if verbosity > 0:
            logger.info(msg)
        if move_local_files:
            try:
                os.rename(src_path, dest_path)
            except Exception as exc:
                logger.error(
                    f'failed to move unmatched file "{src_path}" '
                    f'to "{dest_path}": {exc}'
                )
                continue

        # reformat paths for S3 file relocations
        name = bad_files_inverse_lookup.get(name, name)
        src_path = os.path.join(s3_prefix, name).replace(os.sep, "/")
        dest_path = os.path.join(s3_prefix, base_path).replace(os.sep, "/")
        s3_moves.append((src_path, dest_path))

    # Move S3 files to their new locations
    logger.info(f"Moving {len(s3_moves)} S3 files to new locations...")
    cmd = 'aws s3 mv "{}" "{}"'
    xml_lookup = {}
    for src, dst in s3_moves:
        _cmd = cmd.format(src, dst)
        if move_remote_files:
            os.system(_cmd)
        elif verbosity > 0:
            logger.info(f"\t{_cmd}")
        xml_lookup[src.split(s3_prefix)[-1]] = dst.split(s3_prefix)[-1]

    # rewrite XML_PATH so tracks point to their new locations
    if write_xml:
        with open(xml_path, encoding="utf-8") as _file:
            soup = BeautifulSoup(_file.read(), "xml")
        loc_prefix = os.path.join(
            "file://localhost", usb_path.strip("/"), "DJ Music", ""
        ).replace(os.sep, "/")
        for track in soup.find_all("TRACK"):
            if not track.get("Location"):
                continue
            loc = unquote(track["Location"]).split(loc_prefix)[-1]
            if any((loc.startswith(name) for name in _user_names)):
                continue
            track["Location"] = quote(
                os.path.join(loc_prefix, xml_lookup[loc]).replace(os.sep, "/")
            )

        with open(
            xml_path, mode="wb", encoding=soup.orignal_encoding
        ) as _file:
            _file.write(soup.prettify("utf-8"))


def fuzz_process(file_name, track_name, threshold):
    """Threaded fuzzy search between local file name and Spotify track.

    Args:
        file_name (str): local file name
        track_name (str): spotify track title - artist
        threshold (int): min fuzz ratio to accept file -> Spotify track match

    Returns:
        tuple: (Spotify track title - artist, fuzz ratio with local file)
    """
    ret = ()
    fuzz_ratio = fuzz.ratio(file_name.lower(), track_name.lower())
    if fuzz_ratio >= threshold:
        ret = (track_name, fuzz_ratio)
    return ret


def make_dirs(dest):
    """Creates non-existent directory.

    Args:
        dest (str): directory to create
    """
    if os.name == "nt":
        cwd = os.getcwd()
        path_parts = dest.split("/")
        root = path_parts[0]
        path_parts = path_parts[1:]
        os.chdir(root)
        for part in path_parts:
            os.makedirs(part, exist_ok=True)
            os.chdir(part)
        os.chdir(cwd)
    else:
        os.makedirs(dest, exist_ok=True)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--structure_data",
        required=True,
        help="""JSON with keys:' \
                "spotify_playlists": map playlist_name -> playlist_id
                "playlist_genres": map playlist name -> genre folder
                "users": map Spotify username -> beatcloud username
                "not_matched_genre_lookup": map remainder track folders -> new
                                            folders
                "bad_files": map badly named files -> corrections
                "ignore": list of files that incorrectly match Spotify tracks
                          above the final fuzz_ratio""",
    )
    parser.add_argument("--config_path", help="path to config.json")
    parser.add_argument(
        "--fuzz_ratio",
        default=80,
        type=float,
        help="minimum fuzz ratio to match Spotify track to local file",
    )
    parser.add_argument(
        "--cache_fuzz_results",
        action="store_true",
        help="cache fuzz matches that are supposedly correct...used to "
        "clean up output as fuzz ratio is progressively lowered",
    )
    parser.add_argument(
        "--move_local_files",
        action="store_true",
        help="actual execute local file moves",
    )
    parser.add_argument(
        "--move_remote_files",
        action="store_true",
        help="rename files in S3 so they're under the proper user and "
        "genre folders",
    )
    parser.add_argument(
        "--write_xml",
        action="store_true",
        help="rewrite XML_PATH to fix locations of tracks",
    )
    parser.add_argument(
        "--verbosity",
        "-v",
        action="count",
        default=0,
        help="logging verbosity",
    )
    args = parser.parse_args()

    if not os.path.exists(args.structure_data):
        raise RuntimeError(
            f"required `--structure_data` JSON config "
            f"'{args.structure_data}' doesn't exist"
        )

    with open(args.structure_data, encoding="utf-8") as _file:
        data = json.load(_file)

    # validate all bad_file lookups have .mp3 extensions
    data["bad_files_inverse_lookup"] = {}
    for key, value in data["bad_files"].items():
        if any((not x.endswith(".mp3") for x in [key, value])):
            raise ValueError(f'"{key, value}" must end with ".mp3"')
        data["bad_files_inverse_lookup"][value] = key

    # temporary buffer to hold local file results for special consideration
    # while fuzzy searching
    _matches = {}

    # standard djtools 'config.json' file (for usb_path, aws_profile, etc.)
    with open(args.config_path, encoding="utf-8") as _file:
        config = json.load(_file)

    if args.move_remote_files:
        os.environ["AWS_PROFILE"] = config.get("aws_profile")

    # cached Spotify API results and `aws s3 ls --recursive` on 'dj/music/'
    cache_path = os.path.join(
        os.path.dirname(__file__), ".cache.json"
    ).replace(os.sep, "/")

    if not os.path.exists(cache_path):
        tracks = get_spotify_tracks(config, data["spotify_playlists"])
        files = get_beatcloud_tracks()
        with open(cache_path, "w", encoding="utf-8") as _file:
            json.dump({"tracks": tracks, "files": files}, _file)
    else:
        with open(cache_path, encoding="utf-8") as _file:
            cache = json.load(_file)
            tracks = cache["tracks"]
            files = cache["files"]
        logger.info(
            f"Retrieved {len(tracks)} tracks and {len(files)} files "
            "from cache."
        )

    # display data['users'] contributions to data['spotify_playlists']
    analyze_tracks(tracks, data["users"])

    # filter `aws s3 ls` result for those that exist at config['usb_path']
    local_files = find_local_files(config["usb_path"], files)

    # optionally display the beatcloud files that aren't present locally
    if args.verbosity > 0:
        unfound = set(files).difference(set(local_files))
        logger.info(f"{len(unfound)} files in S3 not found locally:")
        for _file in unfound:
            logger.info(f"\t{_file}")

    # don't consider any files already in the proper username-based folders
    user_names = {
        os.path.join(x, "").replace(os.sep, "/")
        for x in data["users"].values()
    }
    _ = len(local_files)
    local_files = [
        x
        for x in local_files
        if not any((x.startswith(name) for name in user_names))
    ]
    logger.info(
        f"Ignoring {_ - len(local_files)} files that are already "
        "in the right place"
    )

    # rename files (only if `--move_local_files`), otherwise swaps 'bad' local
    # files in index for the 'good' ones mapped in data['bad_files']
    local_files = fix_files(
        data["bad_files"],
        local_files,
        args.move_local_files,
        config["usb_path"],
        args.verbosity,
    )

    # This operation is meant to be run multiple times with progressively lower
    # 'fuzz_ratio'...each time:
    #    - typos and misformattings should be corrected using data['bad_files']
    #    - correct matches should be allowed to cache in `.fuzz_cache.json`
    #    - incorrect matches must be confirmed unattributable to a user and
    #      added to data['ignore']; incorrect matches that are permitted may
    #      result in a track being moved to a technically incorrect user /
    #      genre folder
    _found_tracks, _not_matched = match_local_files(
        local_files,
        tracks,
        args.fuzz_ratio,
        args.verbosity,
        data["ignore"],
        _matches,
        args.cache_fuzz_results,
    )

    # moves local files, remote files, and rewrites XML
    move_files(
        _found_tracks,
        _not_matched,
        data["users"],
        data["playlist_genres"],
        config["usb_path"],
        args.move_local_files,
        data["not_matched_genre_lookup"],
        args.move_remote_files,
        args.verbosity,
        data["ignore"],
        data["bad_files_inverse_lookup"],
        config["XML_PATH"],
        user_names,
        args.write_xml,
    )
